(* SetPi translation of the original AIF example zebsKeyServer.
 * The model uses a keyring for each user and sets of valid and revoked keys for
 * each combination of user and server (one server in this example).
 * Updating a key requires sending a valid signature of the new public key with
 * the old private key, and updating the sets accordingly on both the client
 * and the server.
 *)

type channel.
type agent.
type skey.
fun pk(skey).
type pkey = pk(skey).

set ring_a:skey. set ring_b:skey.
set valid_a_s:pkey. set valid_b_s:pkey.
set revoked_a_s:pkey. set revoked_b_s:pkey.

fun sign(t1,t2).
reduc checksign(pk(x), sign(x,y)) = y.
reduc getmsg(sign(x,y)) = y.

new ch: channel.
new d: channel [private].

query x_sk:skey; att(x_sk) where pk(x_sk) in valid_a_s.
query x_sk:skey; att(x_sk) where pk(x_sk) in valid_b_s.

(* Out-of-band registration of the initial key. Generate a new key pair, store
 * the public key as valid for the server and the private key inside the
 * client's keyring. *)
let Init(a, ring, valid) =
  lock(ring, valid);
  new sk1: skey;
  let pk1:pkey = pk(sk1) in
  update(sk1 in ring and pk1 in valid);
  out(d, sk1);
  out(ch, pk1);
  unlock(ring, valid).

(* The client takes one of its private keys (in their keyring) and performs an
 * update on the server by signing the new public key with their old secret key.
 * When the public key is revoked by the server, also output the old secret
 * key. *)
let Client(a, ring, revoked) =
  lock(ring);
  new sk2:skey;
  in(d, sk1:skey);
  if sk1 in ring then
  let pk2:pkey = pk(sk2) in
  update((sk2 in ring) and (not sk1 in ring));
  out(ch, sign(sk1, (a, pk2)));
  unlock(ring);
  (* Reveal revoked keys *)
  lock(revoked);
  let pk1:pkey = pk(sk1) in
  if pk1 in revoked then
  out(ch, sk1);
  unlock(revoked).

(* Receive an update message from the client, retrieve a matching valid public
 * key, revoke the old key and mark the new key as the valid one. *)
let Server(a, valid, revoked) =
  lock(valid, revoked);
  in(ch, pk1:pkey);
  if pk1 in valid then
  in(ch, x_msg:sign(skey, (agent, pkey)));
  let (=a, x_pk2:pkey) = checksign(pk1, x_msg) in
  if not x_pk2 in revoked then
  update((not pk1 in valid)
         and (pk1 in revoked)
         and (x_pk2 in valid));
  unlock(valid, revoked).

process
  new a: agent; new b: agent;
  (Init(a, ring_a, valid_a_s)) |
  (Init(b, ring_b, valid_b_s)) |
  (!Client(a, ring_a, revoked_a_s)) |
  (!Client(b, ring_b, revoked_b_s)) |
  (!Server(a, valid_a_s, revoked_a_s)) |
  (!Server(b, valid_b_s, revoked_b_s))
